#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Test neutral file generated with NX NASTRAN
"""

import os
import tempfile

import pytest
from click.testing import CliRunner

from femap_neutral_parser import Parser, blocks

from .test_utils import datadir, tempdir  # fixtrures


@pytest.fixture(scope="module")
def neutral(datadir):
    fpath = os.path.join(datadir, "FEMAP_v8-2.neu")
    return Parser(fpath)


def test_basics(neutral):
    assert neutral.femap_version[:6] == "8.1999"
    assert neutral.db_title == ""
    # test table of content
    assert neutral._toc == {100: [[2, 4]], 450: [[7, 29]], 451: [[32, 3550]]}


# =============================================================================
# Block 450 {{1
# =============================================================================
def test_b450_raw_data(neutral):
    """test Block 450 parsing. Since several Blocks 450 are usually found,
    `_get_block_raw_data` yields a file-like objecy containing each of this
    block
    """
    expected = [
        [
            "1,\n",
            "NASTRAN SPC 1 - NASTRAN SPC 1 - NASTRAN SPC 1 - LC1\n",
            "36,1,\n",
            "0.,\n",
            "6,\n",
            "From: Z:\\13-Projects\\10-Air France\\Seat track "
            "analysis\\7-Floortool\\test\\TC_03-000.op2\n",
            "Date : Mon Apr  5 09:56:49 2021\n",
            "<NULL>\n",
            "ANALYSE\n",
            "NASTRAN SPC 1 - NASTRAN SPC 1 - NASTRAN SPC 1 - LC1\n",
            "TEST\n",
            "2,\n",
            "NASTRAN SPC 1 - NASTRAN SPC 1 - NASTRAN SPC 1 - LC2\n",
            "36,1,\n",
            "0.,\n",
            "6,\n",
            "From: Z:\\13-Projects\\10-Air France\\Seat track "
            "analysis\\7-Floortool\\test\\TC_03-000.op2\n",
            "Date : Mon Apr  5 09:56:49 2021\n",
            "<NULL>\n",
            "ANALYSE\n",
            "NASTRAN SPC 1 - NASTRAN SPC 1 - NASTRAN SPC 1 - LC2\n",
            "TEST\n",
        ]
    ]

    # =========================================================================
    # test raw parsing
    # =========================================================================
    b450s = neutral._get_block_raw_data(450, as_file_like=False)
    b450s = list(b450s)
    for i, b450 in enumerate(b450s):
        # b450 = b450.readlines()
        assert b450 == expected[i]

    # =========================================================================
    # also test file-like parsing
    # =========================================================================
    b450s = neutral._get_block_raw_data(450, as_file_like=True)
    for i, b450 in enumerate(b450s):
        b450 = b450.readlines()
        assert b450 == expected[i]


def test_b450_data(neutral):
    """test Block 450 parsing."""
    expected = [
        {
            "ID": 1,
            "anal_type": 1,
            "from_prog": 36,
            "notes": "From: Z:\\13-Projects\\10-Air France\\Seat track "
            "analysis\\7-Floortool\\test\\TC_03-000.op2\n"
            "Date : Mon Apr  5 09:56:49 2021\n"
            "\n"
            "ANALYSE\n"
            "NASTRAN SPC 1 - NASTRAN SPC 1 - NASTRAN SPC 1 - LC1\n"
            "TEST",
            "title": "NASTRAN SPC 1 - NASTRAN SPC 1 - NASTRAN SPC 1 - LC1",
            "value": 0.0,
        },
        {
            "ID": 2,
            "anal_type": 1,
            "from_prog": 36,
            "notes": "From: Z:\\13-Projects\\10-Air France\\Seat track "
            "analysis\\7-Floortool\\test\\TC_03-000.op2\n"
            "Date : Mon Apr  5 09:56:49 2021\n"
            "\n"
            "ANALYSE\n"
            "NASTRAN SPC 1 - NASTRAN SPC 1 - NASTRAN SPC 1 - LC2\n"
            "TEST",
            "title": "NASTRAN SPC 1 - NASTRAN SPC 1 - NASTRAN SPC 1 - LC2",
            "value": 0.0,
        },
    ]

    parsed_block = neutral._parse_blocks(450)
    assert parsed_block.data == expected



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Test neutral file generated with MYSTRAN
"""

import os
import tempfile

import numpy as np
import pytest
from click.testing import CliRunner

from femap_neutral_parser import Parser, blocks

from .test_utils import datadir, tempdir  # fixtrures


@pytest.fixture(scope="module")
def neutral(datadir):
    fpath = os.path.join(datadir, "mystran_00.NEU")
    return Parser(fpath)


@pytest.fixture(scope="module")
def f_neutral(datadir):
    fpath = os.path.join(datadir, "FEMAP_v8-2.neu")
    return Parser(fpath)


def test_basics(neutral):
    assert neutral.header["femap_version"] == "8.2"
    assert neutral.header["db_title"] == "323152540, bulk_model.F06"
    assert neutral.femap_version == "8.2"
    assert neutral.db_title == "323152540, bulk_model.F06"
    # test table of content
    assert neutral._toc == {
        100: [[3, 5]],
        450: [[8, 13], [907, 912]],
        451: [[16, 904], [915, 1598]],
    }
    # check available blocks
    assert neutral.available_blocks() == {
        "header": 100,
        "output_sets": 450,
        "_output_vectors": 451,
    }


# =============================================================================
# Block 450 {{1
# =============================================================================
def test_b450_raw_data(neutral):
    """test Block 450 parsing. Since several Blocks 450 are usually found,
    `_get_block_raw_data` yields a file-like objecy containing each of this
    block
    """
    expected = [
        [
            "       1,\n",
            "Analyse. NASTRAN SPC 1 - lc1. test,          \n",
            "0,1,\n",
            "0.,\n",
            "0,\n",
        ],
        [
            "       2,\n",
            "Analyse. NASTRAN SPC 1 - lc2. test,     \n",
            "0,1,\n",
            "0.,\n",
            "0,\n",
        ],
    ]

    # =========================================================================
    # test raw parsing
    # =========================================================================
    b450s = neutral._get_block_raw_data(450, as_file_like=False)
    for i, b450 in enumerate(b450s):
        # b450 = b450.readlines()
        assert b450 == expected[i]
    # =========================================================================
    # also test file-like parsing
    # =========================================================================
    b450s = neutral._get_block_raw_data(450, as_file_like=True)
    for i, b450 in enumerate(b450s):
        b450 = b450.readlines()
        assert b450 == expected[i]


def test_b450_data(neutral):
    """test Block 450 parsing."""
    expected = [
        {
            "ID": 1,
            "anal_type": 1,
            "from_prog": 0,
            "notes": "",
            "title": "Analyse. NASTRAN SPC 1 - lc1. test",
            "value": 0.0,
        },
        {
            "ID": 2,
            "anal_type": 1,
            "from_prog": 0,
            "notes": "",
            "title": "Analyse. NASTRAN SPC 1 - lc2. test",
            "value": 0.0,
        },
    ]
    parsed_block = neutral._parse_blocks(450)
    assert parsed_block.data == expected


def test_b450_get(neutral):
    """test Block 450 parsing."""
    expected = {
        1: {
            "anal_type": "Static",
            "from_prog": "Unknown",
            "integer_format": None,
            "notes": "",
            "process_type": None,
            "title": "Analyse. NASTRAN SPC 1 - lc1. test",
            "value": 0.0,
        },
        2: {
            "anal_type": "Static",
            "from_prog": "Unknown",
            "integer_format": None,
            "notes": "",
            "process_type": None,
            "title": "Analyse. NASTRAN SPC 1 - lc2. test",
            "value": 0.0,
        },
    }

    assert neutral.output_sets == expected


# =============================================================================
# Block 451 {{1
# =============================================================================


def test_b451_data(neutral):
    """test Block 451 parsing."""
    expected = [
        {
            "ID": 1,
            "anal_type": 1,
            "from_prog": 0,
            "notes": "",
            "title": "Analyse. NASTRAN SPC 1 - lc1. test",
            "value": 0.0,
        },
        {
            "ID": 2,
            "anal_type": 1,
            "from_prog": 0,
            "notes": "",
            "title": "Analyse. NASTRAN SPC 1 - lc2. test",
            "value": 0.0,
        },
    ]
    parsed_block = neutral._parse_blocks(451)
    data = parsed_block.data
    assert len(data) == 84
    assert set(data[0].keys()) == {
        "abs_max",
        "calc_warn",
        "cent_total",
        "comp_dir",
        "component_vec",
        "ent_type",
        "id_max",
        "id_min",
        "max_val",
        "min_val",
        "out_type",
        "record",
        "setID",
        "title",
        "vecID",
    }

    assert data[0]["vecID"] == 10001
    assert data[1]["vecID"] == 10002
    # ...
    assert data[-1]["vecID"] == 50212
    # -------------------------------------------------------------------------
    # full check for last vector
    expected = {
        "abs_max": 45.04958,
        "calc_warn": True,
        "cent_total": True,
        "comp_dir": 3,
        "component_vec": [50201.0, 50202.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        "ent_type": 8,
        "id_max": 1,
        "id_min": 4,
        "max_val": 45.04958,
        "min_val": -2.221523,
        "out_type": 3,
        "setID": 2,
        "title": "BAR EndB Plane1 Moment",
        "vecID": 50202,
    }
    # we drop the array frm data structure since it requires a specific test
    act_without_array = data[-11].copy()
    actual_records = act_without_array.pop("record")
    assert act_without_array == expected
    expected_records = np.array(
        [
            (1, 4.504958e01),
            (2, 5.241556e-01),
            (3, 1.331676e01),
            (4, -2.221523e00),
            (5, 1.318767e-11),
            (6, -1.419975e-14),
            (9, -2.878763e-01),
            (10, 6.273498e-01),
            (11, 5.241556e-01),
        ],
        dtype=[("entityID", "<i8"), ("value", "<f8")],
    )
    np.testing.assert_array_equal(actual_records, expected_records)


def test_b451_get(neutral, f_neutral):
    """test Block 451 parsing."""
    expected = {
        1: {
            "anal_type": "Static",
            "from_prog": "Unknown",
            "integer_format": None,
            "notes": "",
            "process_type": None,
            "title": "Analyse. NASTRAN SPC 1 - lc1. test",
            "value": 0.0,
        },
        2: {
            "anal_type": "Static",
            "from_prog": "Unknown",
            "integer_format": None,
            "notes": "",
            "process_type": None,
            "title": "Analyse. NASTRAN SPC 1 - lc2. test",
            "value": 0.0,
        },
    }
    # -------------------------------------------------------------------------
    # check lazy loading
    assert "_output_vectors" not in dir(neutral)  # NOT loaded
    rss_translation_LCID2 = neutral._output_vectors["displacements::t_total"][2].copy()
    assert "_output_vectors" in dir(neutral)  # loaded
    values = rss_translation_LCID2.pop("record")
    assert rss_translation_LCID2 == {
        "abs_max": 2.578386,
        "calc_warn": True,
        "cent_total": True,
        "comp_dir": 1,
        "component_vec": [10002.0, 10003.0, 10004.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
        "id_max": 7,
        "id_min": 1,
        "max_val": 2.578386,
        "min_val": 0.0,
        "vecID": 10001,
    }
    expected_values = np.array(
        [
            (1, 0.000000e00),
            (2, 2.045391e-01),
            (3, 0.000000e00),
            (4, 1.468270e-02),
            (5, 9.231050e-05),
            (6, 6.276400e-01),
            (7, 2.578386e00),
            (8, 1.025100e-01),
            (9, 2.578363e00),
            (10, 1.916094e00),
            (11, 1.100510e00),
            (12, 2.389742e00),
        ],
        dtype=[("NodeID", "<i8"), ("displacements::t_total", "<f8")],
    )
    np.testing.assert_array_equal(values, expected_values)
    # also check initial key
    # rss_translation_LCID2 = neutral._output_vectors["RSS translation"][2].copy()


def test_vectors_agg(neutral, f_neutral):
    expected = np.array(
        [
            (1, 0.0, 0.0, 0.0000000e00, 1),
            (2, -0.1870816, 0.0, 0.0000000e00, 1),
            (3, 0.0, 0.0, 0.0000000e00, 1),
            (4, 0.07011177, 0.0, 0.0000000e00, 1),
            (5, 0.0, 0.0, -5.7721190e-03, 1),
            (6, 0.0, 0.0, -5.6935130e-01, 1),
            (7, 0.0, 0.0, -1.2992963e00, 1),
            (8, -0.0937607, 0.0, -5.7549970e-03, 1),
            (9, 0.03555086, 0.0, -1.2992791e00, 1),
            (10, 0.0, 0.0, -1.5645016e00, 1),
            (11, 0.0, 0.0, -9.5607290e-01, 1),
            (12, 0.0, 0.0, -1.6029116e00, 1),
            (1, 0.0, 0.0, 0.0000000e00, 2),
            (2, -0.2045391, 0.0, 0.0000000e00, 2),
            (3, 0.0, 0.0, 0.0000000e00, 2),
            (4, -0.0146827, 0.0, 0.0000000e00, 2),
            (5, 0.0, 0.0, -9.2310500e-05, 2),
            (6, 0.0, 0.0, -6.2763990e-01, 2),
            (7, 0.0, 0.0, -2.5783863e00, 2),
            (8, -0.1025099, 0.0, -9.2036680e-05, 2),
            (9, -0.00744509, 0.0, -2.5783525e00, 2),
            (10, 0.0, 0.0, -1.9160941e00, 2),
            (11, 0.0, 0.0, -1.1005101e00, 2),
            (12, 0.0, 0.0, -2.3897419e00, 2),
        ],
        dtype=[
            ("NodeID", "<i8"),
            ("displacements::t1", "<f8"),
            ("displacements::t2", "<f8"),
            ("displacements::t3", "<f8"),
            ("SubcaseID", "<i8"),
        ],
    )
    expected = np.rec.array(expected)
    vec = neutral.vectors(
        ("displacements::t1", "displacements::t2", "displacements::t3")
    )
    f_vec = neutral.vectors(
        ("displacements::t1", "displacements::t2", "displacements::t3")
    )
    np.testing.assert_array_almost_equal(f_vec.view(None), expected.view(None))
    np.testing.assert_array_almost_equal(vec.view(None), expected.view(None))


def test_info(neutral):
    info = neutral.info(doprint=False)
    expected = [
        "\nAnalysis",
        "========",
        " * subcase 1: Analyse. NASTRAN SPC 1 - lc1. test (MYSTRAN::Static)",
        " * subcase 2: Analyse. NASTRAN SPC 1 - lc2. test (MYSTRAN::Static)",
        "\nOutputs",
        "=======",
        "access to one of them using `._output_vectors[<title>][<SubcaseID>]['record']\n",
        " * displacements::t_total -> RSS translation",
        " * displacements::t1 -> T1 translation",
        " * displacements::t2 -> T2 translation",
        " * displacements::t3 -> T3 translation",
        " * displacements::r_total -> RSS rotation",
        " * displacements::r1 -> R1 rotation",
        " * displacements::r2 -> R2 rotation",
        " * displacements::r3 -> R3 rotation",
        " * force_vectors::t_total -> RSS applied force",
        " * force_vectors::t1 -> T1 applied force",
        " * force_vectors::t2 -> T2 applied force",
        " * force_vectors::t3 -> T3 applied force",
        " * force_vectors::r_total -> RSS applied moment",
        " * force_vectors::r1 -> R1 applied moment",
        " * force_vectors::r2 -> R2 applied moment",
        " * force_vectors::r3 -> R3 applied moment",
        " * spc_forces::t_total -> RSS SPC force",
        " * spc_forces::t1 -> T1 SPC force",
        " * spc_forces::t2 -> T2 SPC force",
        " * spc_forces::t3 -> T3 SPC force",
        " * spc_forces::r_total -> RSS SPC moment",
        " * spc_forces::r1 -> R1 SPC moment",
        " * spc_forces::r2 -> R2 SPC moment",
        " * spc_forces::r3 -> R3 SPC moment",
        " * cbar_force::bending_moment_a1 -> BAR EndA Plane1 Moment",
        " * cbar_force::bending_moment_b1 -> BAR EndB Plane1 Moment",
        " * cbar_force::bending_moment_a2 -> BAR EndA Plane2 Moment",
        " * cbar_force::bending_moment_b2 -> BAR EndB Plane2 Moment",
        " * cbar_force::shear1 -> BAR EndB Pl1 Shear Force",
        " * cbar_force::shear2 -> BAR EndB Pl2 Shear Force",
        " * cbar_force::axial -> BAR EndA Axial Force",
        " * cbar_force::torque -> BAR EndA Torque",
        " * cbar_stress::comb_a1 -> BAR EndA Pt1 Comb Stress",
        " * cbar_stress::comb_b1 -> BAR EndB Pt1 Comb Stress",
        " * cbar_stress::comb_a2 -> BAR EndA Pt2 Comb Stress",
        " * cbar_stress::comb_b2 -> BAR EndB Pt2 Comb Stress",
        " * cbar_stress::comb_a3 -> BAR EndA Pt3 Comb Stress",
        " * cbar_stress::comb_b3 -> BAR EndB Pt3 Comb Stress",
        " * cbar_stress::comb_a4 -> BAR EndA Pt4 Comb Stress",
        " * cbar_stress::comb_b4 -> BAR EndB Pt4 Comb Stress",
        " * cbar_stress::max_comb_a -> BAR EndA Max Stress",
        " * cbar_stress::max_comb_b -> BAR EndB Max Stress",
        " * cbar_stress::min_comb_a -> BAR EndA Min Stress",
        " * cbar_stress::min_comb_b -> BAR EndB Min Stress",
    ]
    assert info == expected


def test_get_local_forces(neutral):
    df = neutral.get("cbar_force", asdf=True)



"""
This module contains defined blocks. Any block shall inherit from `Block`.
"""

import logging
from collections import defaultdict

from femap_neutral_parser.blocks._base import Block
from femap_neutral_parser.utils import CaseInsensitiveDict

# =============================================================================
# Block B451
# =============================================================================


class B451(Block):
    """
    >>> txt = '''       1,   10001,       1
    ... RSS translation
    ...      0.000000E+00,     1.602912E+00,     1.602912E+00,
    ...    10002,   10003,   10004,       0,       0,       0,       0,       0,       0,       0,
    ...        0,       0,       0,       0,       0,       0,       0,       0,       0,       0,
    ...        1,      12,       1,       7,
    ...        1,       1,       1
    ...        1,     0.000000E+00,
    ...        2,     1.870816E-01,
    ...        3,     0.000000E+00,
    ...        4,     7.011176E-02,
    ...       -1,     0.          ,
    ... '''
    >>> b = B451(version=9.3)
    >>> b.parse(txt)
    >>> __import__('pprint').pprint(b.data)
    [{'abs_max': 1.602912,
      'calc_warn': True,
      'cent_total': True,
      'comp_dir': 1,
      'component_vec': [10002.0,
                        10003.0,
                        10004.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0],
      'ent_type': 7,
      'id_max': 12,
      'id_min': 1,
      'max_val': 1.602912,
      'min_val': 0.0,
      'out_type': 1,
      'record': array([(1, 0.        ), (2, 0.1870816 ), (3, 0.        ), (4, 0.07011176)],
          dtype=[('entityID', '<i8'), ('value', '<f8')]),
      'setID': 1,
      'title': 'RSS translation',
      'vecID': 10001}]
    """

    NAME = "_output_vectors"
    RECORDS = [
        [
            {"field": "setID", "coerce": int},
            {"field": "vecID", "coerce": int},
            {"field": "_", "coerce": int},
        ],  # 1
        [
            {"field": "title"},
        ],  # 2
        [
            {"field": "min_val", "coerce": float},  # 3
            {"field": "max_val", "coerce": float},
            {"field": "abs_max", "coerce": float},
        ],
        [  # 4 comp[0..9]
            {"field": "component_vec", "coerce": float, "single_line_array": 10},
        ],
        [  # 5 comp[10..19]
            {"field": "_", "coerce": float, "single_line_array": 10},
        ],
        [  # 6
            {"field": "doubled_sided_contour", "coerce": int, "v+": "10.0"},
        ],
        [  # 7
            {"field": "id_min", "coerce": int},
            {"field": "id_max", "coerce": int},
            {"field": "out_type", "coerce": int},
            {"field": "ent_type", "coerce": int},
        ],
        [  # 8
            {"field": "calc_warn", "coerce": bool},
            {"field": "comp_dir", "coerce": int, "v+": "4.1"},
            {"field": "cent_total", "coerce": bool},
        ],
        [  # 1 record for each entry, plus last record
            {
                "multi_line_array": True,
                "field": "record",
                "fields": ("entityID", "value"),
                "coerce": (int, float),
            },
        ],
    ]

    def digest(self, translate):
        agg = defaultdict(dict)  # returned dictionnary
        for data in self.data:
            # -----------------------------------------------------------------
            # affect appropriate columns headers
            ent_type = {7: "NodeID", 8: "ElementID"}[data.pop("ent_type")]
            out_type = {
                0: "value",
                1: "disp",
                2: "accel",
                3: "force",
                4: "stress",
                5: "strain",
                6: "temp",
            }.get(data.pop("out_type"), "user")
            # -----------------------------------------------------------------
            # modify MYSTRAN/FEMAP titles removing duplicated blanks
            title = " ".join(data.pop("title").split())
            try:
                # neutral TITLE -> (<vector>, <axis>)
                vector, new_title = translate(key=title, merge_aliases=True)
            except ValueError:
                # this is an unkown vector
                vector, new_title = "?", title
            else:
                new_title = "::".join((vector, new_title))
            data["record"].dtype.names = (ent_type, new_title)
            set_id = data.pop("setID")
            agg[new_title][set_id] = data
            # agg[title][set_id] = data  # also store initial keys
        return CaseInsensitiveDict(agg)


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)
    # import sys
    # filepath =sys.argv[1]



import logging
from abc import ABC, abstractmethod
from copy import deepcopy
from io import StringIO
from pprint import pformat as pp
from typing import Iterable, List, Tuple, Union

import numpy as np


def version_ge(v1: str, v2: str) -> bool:
    """
    Check a version against another.

    return True if `v1` > `v2` (`v1` more recent than `v2`),
    or if `v2` is 0 or None.

    This function is very naive for now, just comparing v1 and v2 as float.

    Maybe use semver instead (https://pypi.org/project/semver/)?

    >>> version_ge("8.2", "9.1")
    False
    >>> version_ge("8.3", "")
    True
    """
    if not v2:
        return True
    return float(v1) >= float(v2)


def read_array(
    content: List[str], line_no: int, fields: List[str], coerce: List[type]
) -> np.ndarray:
    """
    extract and turn a bunch of lines into a numpy structured array (by default).

    >>> txt = '''line 1,
    ... line 2,
    ... line 3, bla, bla,
    ... 1, 10,
    ... 2, 20,
    ... 3, 30,
    ... 4, -5
    ... -1, 0
    ... bla, bla,
    ... '''
    >>> txt = txt.split('\\n')
    >>> data = read_array(txt, line_no=3, fields=('col1', 'col2'), coerce=(int, float))
    >>> data
    array([(1, 10.), (2, 20.), (3, 30.), (4, -5.)],
          dtype=[('col1', '<i8'), ('col2', '<f8')])

    which can be easily turned to pandas df:

    >>> import pandas as pd
    >>> df = pd.DataFrame(data)
    >>> print(df)
       col1  col2
    0     1  10.0
    1     2  20.0
    2     3  30.0
    3     4  -5.0

    """
    read_lines = 0
    data = []
    for line in content[line_no:]:
        if line.strip().startswith("-1"):
            break
        data.append(tuple(line.strip().strip(",").split(",")))
    return np.array(data, dtype=list(zip(fields, coerce)))


# =============================================================================
# blocks
# =============================================================================
class Block(ABC):
    """
    Abstract Base Class for Neutral blocks.

    Mandatory methods to ovverride are:
        * `digest`: a classmethod converting a list of structured parsed data
        into the expected end-user data structure.
    """

    RECORDS = []  # to be overwritten by children

    def __init__(self, version: str):
        # clean specs and remove fields when current version doesn't match
        self._records = deepcopy(self.RECORDS)
        for rno, records in enumerate(self.RECORDS):
            _records = [r for r in records if version_ge(version, r.get("v+"))]
            self._records[rno] = _records
        # clen empty mist
        self._records = [s for s in self._records if s]
        # from pprint import pprint as pp
        # pp(self.RECORDS)
        self.data = []
        self.flavor = None

    @classmethod
    def id(cls):
        return int(cls.__name__[1:])

    def parse(self, content: Union[StringIO, str, Iterable[str]], debug=False) -> None:
        """parse provided `content`.
        `content`may be either:
            * a file-like object
            * a single multi-line string
            * an iterable (list or tuple) of strings
        """
        if debug:
            breakpoint()
        # =====================================================================
        # ensure content is a list of lines
        # =====================================================================
        if isinstance(content, StringIO):
            content = [l.rstrip("\n") for l in content.readlines()]
        elif isinstance(content, str):
            content = content.split("\n")
        else:
            content = [l.rstrip("\n") for l in content]
        # drop empty strings
        content = [e for e in content if e]
        # =====================================================================
        # loop control variables
        # because blocks can span several lines...
        # =====================================================================
        next_lines_to_read = 1
        lines_offset = 0  # in case some loose lines exist like eg. block450 "notes"
        # ---------------------------------------------------------------------
        # one block [-1 ... -1] can be made of several blocks
        # without the -1 / -1 signal!
        _content_loop = 0  # debugging-only purposes
        while content:
            data = {}
            for fields_set_no, fields_set in enumerate(self._records):
                # -----------------------------------------------------------------
                # for each set of fields (list of dictionnaries)
                line_no = lines_offset + fields_set_no
                lines = "\n".join(
                    content[line_no : line_no + next_lines_to_read]
                ).replace("<NULL>", "")
                lines_offset += next_lines_to_read - 1
                # -----------------------------------------------------------------
                # if single-line multi-fields, split fields by coma
                # this may be reveresed for `comp` fields later on
                if next_lines_to_read == 1:
                    fields = [f.strip() for f in lines.strip(",").split(",")]
                # -----------------------------------------------------------------
                # if multi-line, single-field, just wrap the lines into a list
                else:
                    fields = [lines]
                for field_no, field_specs in enumerate(fields_set):
                    fieldname = field_specs.get("field")
                    coerce_func = field_specs.get("coerce", lambda x: x)
                    # -------------------------------------------------------------
                    # read multi-line array
                    if field_specs.get("multi_line_array", False):
                        field_specs.get("multi_line_array")
                        records = read_array(
                            content,
                            line_no,
                            field_specs["fields"],
                            field_specs["coerce"],
                        )
                        lines_offset += len(records)
                    # -------------------------------------------------------------
                    # read single-line array
                    elif field_specs.get("single_line_array", False):
                        # one field wraps an array
                        nbi = field_specs["single_line_array"]
                        records = [coerce_func(e) for e in fields[:nbi]]
                    # -------------------------------------------------------------
                    # single-line, single-field
                    elif field_specs.get("one_field", False):
                        # rejoin previously splitted line
                        records = ", ".join(fields)
                    # -------------------------------------------------------------
                    # single-line, multi-fields
                    else:
                        try:
                            field = coerce_func(fields[field_no])
                        except Exception as exc:
                            logging.critical(
                                f"loop#{_content_loop}:: cannot coerce field '{fieldname}' ({fields[field_no]}) with {coerce_func}"
                            )
                            logging.critical(f"{lines_offset=}")
                            logging.critical(f"{fields_set=}")
                            logging.critical(f"content={pp(content)}, {fields=}")
                            logging.critical(f"{lines=}")
                            raise
                        # if current field describe nb of next lines to read:
                        if field_specs.get("is_offset", False):
                            next_lines_to_read = field
                            continue
                        else:
                            next_lines_to_read = 1
                            records = field
                    data[fieldname] = records
            else:
                # last_line = lines_offset + fields_set_no
                # _parsed_content = content[:last_line]
                # if self.__class__.__name__ == "B451":
                #     breakpoint()
                content = content[lines_offset + fields_set_no + 1 :]
                lines_offset = 0
                # remove dummy data:
                if "_" in data:
                    del data["_"]
                self.data.append(data)
                _content_loop += 1
            # return data

    @abstractmethod
    def digest(self):
        pass

    def _clean_default(self, txt):
        return txt.strip().strip(",").strip()


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)



import logging
import os
from collections import defaultdict as dd
from functools import reduce
from io import StringIO
from typing import Dict, List, Tuple

import numpy as np
import numpy.lib.recfunctions as rfn

from femap_neutral_parser import blocks
from femap_neutral_parser.neutral_flavors import Flavor

try:
    import pandas as pd

    ISPANDAS = True
except ImportError:
    ISPANDAS = False


def _todf(array):
    if not ISPANDAS:
        raise ValueError("Pandas library is not available")
    df = pd.DataFrame(array)
    ix = [c for c in ["SubcaseID", "NodeID", "ElementID"] if c in df]
    return df.set_index(ix)


def get_flavor(from_prog):
    if from_prog == "Unknown":
        return "MYSTRAN"
    return from_prog


class Parser(Flavor):
    def __init__(self, fpath: str) -> None:
        self.fpath = os.path.expanduser(fpath)
        if not os.path.isfile(self.fpath):
            raise ValueError(f'file path "{fpath}" not found')
        self._build_toc()  # -> self._toc
        self._parse_header()
        # self.blocks = {}
        # ---------------------------------------------------------------------
        # collect available defined blocks (code-wise, **not** neutral file wise)
        # this will create once self.__getattr__ will be called:
        #  * `self.header` from B100
        #  * `self.output_sets` from B450
        #  * `self._output_vectors` from B451
        self._defined_block_names = {
            getattr(blocks, b).NAME: getattr(blocks, b)
            for b in dir(blocks)
            if b.startswith("B")
        }
        # get flavor
        self.flavor = get_flavor(
            self.output_sets[list(self.output_sets.keys())[0]]["from_prog"]
        )
        logging.info(f"flavor={self.flavor}")

    def info(self, doprint=True, show_unknown=True):
        msg = []
        if 450 in self._toc:
            msg.append("\nAnalysis")
            msg.append("========")
            for lcid, lc_data in self.output_sets.items():
                flavor = get_flavor(lc_data["from_prog"])
                msg.append(
                    f" * subcase {lcid}: {lc_data['title']} ({flavor}::{lc_data['anal_type']})"
                )
        if 451 in self._toc:
            msg.append("\nOutputs")
            msg.append("=======")
            msg.append(
                "access to one of them using `._output_vectors[<title>][<SubcaseID>]['record']\n"
            )
            for output_title, output_data in self._output_vectors.items():
                # lcids = ", ".join(map(str, output_data.keys()))
                try:
                    initial_title = self.headers_to_title(output_title)
                    _field = f" * {output_title}"
                    isknown = True
                except ValueError:
                    # logging.warning(f"{output_title} is not known")
                    initial_title = f"{output_title}"
                    _field = f" * ## {output_title}"
                    isknown = False
                if not isknown and not show_unknown:
                    # skip Unknown vectors
                    continue
                _field += f" -> {initial_title}"
                msg.append(_field)
        if doprint:
            print("\n".join(msg))
            return
        return msg

    def vectors(
        self,
        titles,
        SubcaseIDs=None,
        raw=False,
        asdf=False,
        prefixed=True,
        incl_totals=True,
    ):
        if not SubcaseIDs:
            SubcaseIDs = tuple(self.output_sets.keys())
        if isinstance(SubcaseIDs, int):
            SubcaseIDs = (SubcaseIDs,)
        if isinstance(titles, str):
            titles = (titles,)
        recs = []
        key = None
        for SubcaseID in SubcaseIDs:
            arrays = []
            for title in titles:
                try:
                    ov = self._output_vectors[title.lower()]
                except KeyError:
                    raise KeyError(f"vector {title} is unknown")
                try:
                    data = ov[SubcaseID]
                except KeyError:
                    # no results for this loadcase
                    logging.warning(f"no results for {SubcaseID} / {title}")
                    continue
                arrays.append(data["record"])
            if not key:
                key = "NodeID" if "NodeID" in arrays[0].dtype.names else "ElementID"
            arr = reduce(lambda a1, a2: rfn.rec_join(key, a1, a2), arrays)
            arr = rfn.append_fields(
                arr, "SubcaseID", [SubcaseID] * len(arr), asrecarray=True, usemask=False
            )
            recs.append(arr)
        arr = rfn.stack_arrays(recs, asrecarray=True, usemask=False)
        return self._clean_df_goodies(
            arr,
            neutral_headers=titles,
            raw=raw,
            asdf=asdf,
            prefixed=prefixed,
            incl_totals=incl_totals,
        )

    def available_blocks(self) -> Dict[int, str]:
        return {
            getattr(blocks, b).NAME: getattr(blocks, b).id()
            for b in dir(blocks)
            if b.startswith("B")
        }

    def __getattr__(self, key: str):
        # look for blocks having key as NAME
        try:
            block_id = self._defined_block_names[key].id()
        except KeyError as exc:
            selfattrs = [attr for attr in dir(self) if attr.startswith("get_")]
            msg = f"'{key}' not found in _defined_block_names."
            msg += f"\n  available attributes are {selfattrs}"
            raise KeyError(msg)
        block = self._parse_blocks(block_id)
        if block_id == 451:
            data = block.digest(translate=self.title_to_headers)
        else:
            data = block.digest()
        setattr(self, key, data)
        return data

    def _parse_header(self):
        """parse block 100. This needs to be done at low-level since we will
        pick FEMAP's version number from there."""
        txt = tuple(self._get_block_raw_data(100))[0]
        b100 = blocks.B100(version=0)  # we do not have correct version at this point
        b100.parse(txt)
        self.header = b100.data[0]
        # also assign directly for convenieance
        self.femap_version = b100.data[0]["femap_version"]
        self.db_title = b100.data[0]["db_title"]

    def _build_toc(self, verbose=True):
        """parse the document to retrieve blocks locations"""
        _toc = dd(list)
        nb_lines = 0
        current_block = None
        for line_nb, line in enumerate(open(self.fpath, "r")):
            # if line_nb == 1598:
            #     breakpoint()
            line = line.strip()
            if line == "-1":
                if not current_block:
                    logging.debug(f"{line_nb=}:: {line}, new block is coming")
                    # ---------------------------------------------------------
                    # new block is coming, block ID known the line after
                    # ---------------------------------------------------------
                    current_block = -1  # set waiting status
                    continue
                else:
                    logging.debug(
                        f"{line_nb=}:: {line}, block {current_block} is finishing"
                    )
                    # ---------------------------------------------------------
                    # block is finishing
                    # ---------------------------------------------------------
                    _toc[current_block][-1].append(line_nb)
                    current_block = None
                    continue
            if current_block == -1:
                # -------------------------------------------------------------
                # waiting status
                current_block = int(line)
                logging.debug(
                    f"{line_nb=}:: {line}, block {current_block} is beginning"
                )
                _toc[current_block].append([line_nb + 1])
            nb_lines += 1
        # ---------------------------------------------------------------------
        # EOF
        # we sometimes miss trailing "-1". fix it:
        if current_block is not None and current_block > 0:
            _toc[current_block][-1].append(line_nb)
            logging.debug(
                f"{line_nb=}:: {line}, arbitrarilly closing {current_block} is finishing"
            )
        self._toc = dict(_toc)
        logging.info("parsed %d lines", nb_lines)

    def _get_block_raw_data(self, block_id, as_file_like=False):
        """yield file-like objects containing block data"""
        fh = open(self.fpath)
        line_offset = 0
        for line_start, line_end in self._toc[block_id]:
            lines = range(line_start, line_end)
            lines_content = []
            for line_nb, line in enumerate(fh, start=line_offset):
                if line_nb > line_end:
                    line_offset = line_nb + 1
                    break
                if line_nb in lines:
                    lines_content.append(line)
            if as_file_like:
                txt = StringIO()
                txt.writelines(lines_content)
                txt.seek(0)
            else:
                txt = lines_content
            yield txt
        fh.close()

    def _parse_blocks(self, block_id: int) -> None:
        """instantiate a block and feed it with relevant raw block text.
        Next step is supposed to be block.digest()
        """
        block = getattr(blocks, f"B{block_id}")(version=self.femap_version)
        for txt in self._get_block_raw_data(block_id=block_id):
            block.parse(txt)
        return block

    # =========================================================================
    # engineering shortcuts goodies
    # =========================================================================
    def get(self, what, asdf=False, prefixed=False, incl_totals=False, raw=False):
        try:
            neutral_headers = self.get_vectors(what, merge_aliases=True)
        except KeyError:
            raise KeyError(
                f"{what=} is not implemented. Select one of {set(self.get_vectors())}"
            )
        data = self.vectors(
            neutral_headers,
            raw=raw,
            asdf=asdf,
            prefixed=prefixed,
            incl_totals=incl_totals,
        )
        return data

    def _clean_df_goodies(
        self,
        data,
        neutral_headers,
        raw=False,
        asdf=False,
        prefixed=False,
        incl_totals=False,
    ):
        if not incl_totals:
            data = rfn.drop_fields(
                data, [f for f in rfn.get_names(data.dtype) if "total" in f]
            )
        if not prefixed and not raw:
            # drop vector from "<vector>::<axis>"
            renaming = {
                t: t.split("::")[1] for t in rfn.get_names(data.dtype) if "::" in t
            }
            data = rfn.rename_fields(data, renaming)
        if raw:
            renaming = {t: self.headers_to_title(t) for t in neutral_headers}
            data = rfn.rename_fields(data, renaming)
        if asdf:
            data = _todf(data)
        return data

    # -------------------------------------------------------------------------
    # Bar engineering forces
    def get_bar_local_forces(
        self,
        raw=False,
        asdf=False,
        prefixed=False,
        incl_totals=False,
    ):
        vectorsA = [
            "cbar_force::axial",
            "cbar_force::shear1",
            "cbar_force::shear2",
            "cbar_force::torque",
            "cbar_force::bending_moment_a1",
            "cbar_force::bending_moment_a2",
        ]
        pntA = self.vectors(vectorsA, prefixed=True)
        vectorsB = [
            "cbar_force::axial",
            "cbar_force::shear1",
            "cbar_force::shear2",
            "cbar_force::torque",
            "cbar_force::bending_moment_b1",
            "cbar_force::bending_moment_b2",
        ]
        pntB = self.vectors(vectorsB, prefixed=True)
        # reverse sign for point B
        for col in vectorsB:
            pntB[col] = -1 * pntB[col]
        # ---------------------------------------------------------------------
        # keep track of extremity
        pntA = rfn.append_fields(
            pntA, "ga/gb", ["A"] * len(pntA), asrecarray=True, usemask=False
        )
        pntB = rfn.append_fields(
            pntB, "ga/gb", ["B"] * len(pntA), asrecarray=True, usemask=False
        )
        neutral_headers = self.get_vectors("cbar_force")
        renaming = {
            t: t.split("::")[1].replace("_a", "")
            for t in rfn.get_names(pntA.dtype)
            if "::" in t
        }
        pntA = rfn.rename_fields(pntA, renaming)
        renaming = {
            t: t.split("::")[1].replace("_b", "")
            for t in rfn.get_names(pntB.dtype)
            if "::" in t
        }
        pntB = rfn.rename_fields(pntB, renaming)
        # ---------------------------------------------------------------------
        # stack
        data = rfn.stack_arrays((pntA, pntB), asrecarray=True, usemask=False)
        # ---------------------------------------------------------------------
        # change signs for loads (and not moments)
        for col in ("axial", "shear1", "shear2", "torque", "bending_moment1"):
            data[col] = -1 * data[col]
        data = self._clean_df_goodies(
            data, neutral_headers, raw, asdf, prefixed, incl_totals
        )
        return data


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)



"""
Common mappings for MYSTRAN and FEMAP
for each self.flavor, describe vectors in terms of (<vector>, <axis>)
"""
import json
from collections import defaultdict
from importlib import resources

import pkg_resources

from femap_neutral_parser.utils import CaseInsensitiveDict

with resources.open_text("femap_neutral_parser.data", "neutral_flavors.json") as fh:
    _headers_to_neutral = json.loads(fh.read())
# due to json restrictions, we cannot save it wit(vector, axis) keys
# hence, we need to 'compact' it:
HEADERS_TO_NEUTRAL = {}
for vector, data in _headers_to_neutral.items():
    for axis, titles in data.items():
        HEADERS_TO_NEUTRAL[(vector, axis)] = titles
# =============================================================================
# the following boilerplate is purely cosmetcis and ensure that
# engineering high-level functions will get t1 before r1
# =============================================================================
HEADERS_TO_NEUTRAL = list(HEADERS_TO_NEUTRAL.items())
# tercery key
HEADERS_TO_NEUTRAL = sorted(HEADERS_TO_NEUTRAL, key=lambda x: x[0][1], reverse=False)
# secondary key: sort by <axis> first letter, "t" before "r"
HEADERS_TO_NEUTRAL = sorted(HEADERS_TO_NEUTRAL, key=lambda x: x[0][1][0], reverse=True)
# primary key: sort by <vector>
HEADERS_TO_NEUTRAL = sorted(HEADERS_TO_NEUTRAL, key=lambda x: x[0][0], reverse=False)
# and go back to dict...
HEADERS_TO_NEUTRAL = dict(HEADERS_TO_NEUTRAL)


# ALIASES are used to restrict `get_vectors` returned values
ALIASES = {
    "cbar_force": {
        "axial_b": "axial",
        "axial_a": "axial",
        "torque_b": "torque",
        "torque_a": "torque",
        "shear_a2": "shear2",
        "shear_a1": "shear1",
        "shear_b2": "shear2",
        "shear_b1": "shear1",
    },
    "cbeam_force": {
        "axial_b": "axial",
        "axial_a": "axial",
        "torque_b": "torque",
        "torque_a": "torque",
        "shear_a2": "shear2",
        "shear_a1": "shear1",
        "shear_b2": "shear2",
        "shear_b1": "shear1",
    },
}

REVERTED_ALIASES = {}
for vector, merges in ALIASES.items():
    REVERTED_ALIASES[vector] = {v: k for k, v in merges.items()}


# -----------------------------------------------------------------------------
# prepare reversed search
# "RSS translation" -> ("displacements", "t_total")  # as pynastran
NEUTRAL_TO_HEADERS = CaseInsensitiveDict()
for headers, titles in HEADERS_TO_NEUTRAL.items():
    for title in titles:
        if title:
            if title in NEUTRAL_TO_HEADERS and headers != NEUTRAL_TO_HEADERS[title]:
                raise ValueError(
                    f"CaseInsensitiveDict cannot handle a different {title=}\nAlready stored {NEUTRAL_TO_HEADERS[title]}\nWant to add {headers}"
                )
            NEUTRAL_TO_HEADERS[title] = headers


class Flavor:
    """Mixin handling neutral self.flavor

    >>> f = Flavor()
    >>> f.title_to_headers("T1 translation")
    ('displacements', 't1')
    >>> f.title_to_headers("t1 TrAnslation")
    ('displacements', 't1')
    """

    def title_to_headers(self, key, merge_aliases=False):
        """Convert Neutral title to headers tuple:

        >>> f = Flavor()
        >>> f.title_to_headers("T1 translation")
        ('displacements', 't1')
        >>> f.title_to_headers("BAR EndA Torque")
        ('cbar_force', 'torque_a')
        >>> f.title_to_headers("BAR EndA Torque", merge_aliases=True)
        ('cbar_force', 'torque')
        """
        headers = NEUTRAL_TO_HEADERS.get(key)
        if not headers:
            # return initial FEMAP/MYSTRAN title
            return key
        vector, axis = headers
        if merge_aliases:
            # check if an alias esits
            aliases = ALIASES.get(vector)
            if aliases:
                axis = aliases.get(axis, axis)
        return (vector, axis)

    def headers_to_title(self, key):
        """search for `key` in `self.flavor` dict
        >>> f = Flavor()
        >>> f.headers_to_title(("displacements", "t1"))
        'T1 translation'
        >>> f.headers_to_title("displacements::t1")
        'T1 translation'
        >>> f.headers_to_title(('cbar_force', 'torque_a'))
        'BAR EndA Torque'
        >>> f.headers_to_title(('cbar_force', 'torque_a'))
        'BAR EndA Torque'
        >>> f.headers_to_title(('cbar_force', 'torque'))
        'BAR EndA Torque'
        """
        if isinstance(key, str):
            vector, axis = key.split("::")
        else:
            # assumed a correct tuple was passed
            vector, axis = key
        try:
            axis = REVERTED_ALIASES[vector][axis]
        except KeyError:
            pass  # no alias
        title = HEADERS_TO_NEUTRAL[(vector, axis)][0]
        return title

    @classmethod
    def get_vectors(self, what=None, merge_aliases=False):
        """
        return a list of relevant vectors
        >>> f = Flavor()
        >>> f.get_vectors("cbar_force") == [
        ... 'cbar_force::torque_a',
        ... 'cbar_force::torque_b',
        ... 'cbar_force::shear_a1',
        ... 'cbar_force::shear_a2',
        ... 'cbar_force::shear_b1',
        ... 'cbar_force::shear_b2',
        ... 'cbar_force::bending_moment_a1',
        ... 'cbar_force::bending_moment_a2',
        ... 'cbar_force::bending_moment_b1',
        ... 'cbar_force::bending_moment_b2',
        ... 'cbar_force::axial_a',
        ... 'cbar_force::axial_b']
        True

        If merge_aliases is `True`::
        >>> f.get_vectors("cbar_force", merge_aliases=True) == [
        ... 'cbar_force::bending_moment_a1',
        ... 'cbar_force::bending_moment_a2',
        ... 'cbar_force::bending_moment_b1',
        ... 'cbar_force::bending_moment_b2',
        ... 'cbar_force::axial',
        ... 'cbar_force::shear1',
        ... 'cbar_force::shear2',
        ... 'cbar_force::torque']
        True

        Ommitting key will give access to available vectors families::
        >>> sorted(list(f.get_vectors()))
        ['applied_gpf', 'cbar_force', ..., 'spc_forces', 'summed_gpf']

        """
        dic = defaultdict(list)
        for (vector, axis), initial_title in HEADERS_TO_NEUTRAL.items():
            dic[vector].append(axis)
        dic = dict(dic)
        # -> {..., 'cbar_force': ['torque_a', 'torque_b',...], ...}
        if merge_aliases:
            for vector, merges in ALIASES.items():
                dic[vector] = [axis for axis in dic[vector] if axis not in merges]
                dic[vector] += sorted(list(set(merges.values())))
        if what is not None:
            return [f"{what}::{t}" for t in dic[what]]
        return set(dic.keys())


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS)



"""
This module contains defined blocks. Any block shall inherit from `Block`.
"""

import logging

from femap_neutral_parser.blocks._base import Block

FROM_PROG = {
    0: "Unknown",
    1: "FEMAP Generated",
    2: "PAL",
    3: "PAL 2",
    4: "MSC/NASTRAN",
    5: "ANSYS",
    6: "STARDYNE",
    7: "COSMOS",
    8: "PATRAN",
    9: "FEMAP Neutral",
    10: "ALGOR",
    11: "SSS/NASTRAN",
    12: "Comma Separated",
    13: "UAI/NASTRAN",
    14: "Cosmic NASTRAN",
    15: "STAAD",
    16: "ABAQUS",
    17: "WECAN",
    18: "MTAB/SAP",
    19: "CDA/Sprint",
    20: "CAEFEM",
    21: "I-DEAS",
    22: "ME/NASTRAN",
    26: "CSA/NASTRAN",
    27: "CFDesign",
    28: "LS-DYNA",
    29: "MARC",
    30: "SINDA",
    31: "NE/Nastran",
    32: "SAToolkit",
    33: "TMG",
    34: "FEMAP Structural",
    35: "VR/Nastran",
    36: "NX Nastran",
}

ANAL_TYPE = {
    0: "Unknown",
    1: "Static",
    2: "Modes",
    3: "Transient",
    4: "Frequency Response",
    5: "Response Spectrum",
    6: "Random",
    7: "Linear Buckling",
    8: "Design Opt",
    9: "Explicit",
    10: "Nonlinear Static",
    11: "Nonlinear Buckling",
    12: "Nonlinear Transient",
    19: "Comp Fluid Dynamics",
    20: "Steady State Heat Transfer",
    21: "Transient Heat",
    22: "Advanced Nonlinear Static",
    23: "Advanced Nonlinear Transient",
    24: "Advanced Nonlinear Explicit",
    25: "Static Aeroelasticity",
    26: "Aerodynamic Flutter",
    27: "Multistep Nonlinear",
    28: "Multistep_Kinematic",
    29: "Preload",
}

PROCESS_TYPE = {
    0: "None",
    1: "Linear Combination",
    2: "RSS Combination",
    3: "Max Envelope",
    4: "Min Envelope",
    5: "AbsMax Envelope",
    6: "Max Envelope SetID",
    7: "Min Envelope SetID",
    8: "AbsMax Envelope SetID",
}


class B450(Block):
    """
    >>> txt = '''           1,
    ... Analyse. NASTRAN SPC 1 - lc1. test,
    ... 0,1,
    ... 0.,
    ... 2,
    ... This is a line
    ... and a second one
    ... 1,2,'''
    >>> b = B450(version=11.1)
    >>> b.data
    []
    >>> from pprint import pprint as pp
    >>> b.parse(txt)
    >>> b.data == [{
    ...  'ID': 1,
    ...  'anal_type': 1,
    ...  'attach_id': 1,
    ...  'from_prog': 0,
    ...  'location_id': 2,
    ...  'notes': 'This is a line\\nand a second one',
    ...  'title': 'Analyse. NASTRAN SPC 1 - lc1. test',
    ...  'value': 0.0}]
    True
    >>> b.digest() == {
    ... 1: {'anal_type': 'Static',
    ...          'from_prog': 'Unknown',
    ...          'integer_format': None,
    ...          'notes': 'This is a line\\nand a second one',
    ...          'process_type': None,
    ...          'title': 'Analyse. NASTRAN SPC 1 - lc1. test',
    ...          'value': 0.0}}
    True
    """

    NAME = "output_sets"
    RECORDS = [
        [
            {"field": "ID", "coerce": int},
        ],  # 1
        [
            {"field": "title"},
        ],  # 2
        [
            {"field": "from_prog", "coerce": int},  # 3
            {"field": "anal_type", "coerce": int},
            {"field": "process_type", "coerce": int, "v+": "11.2"},
            {"field": "integer_format", "coerce": bool, "v+": "11.2"},
        ],
        [
            {"field": "value", "coerce": float},
        ],  # 4
        [
            {"field": "nlines", "coerce": int, "is_offset": True},
        ],  # 5
        [
            {"field": "notes"},
        ],  # may be null
        [
            {"field": "attach_id", "coerce": int, "v+": "11.0"},  # 6
            {"field": "location_id", "coerce": int, "v+": "11.0"},
            {"field": "study_id", "coerce": int, "v+": "11.2"},
        ],
    ]

    def digest(self):

        ret = {}
        for dic in self.data:
            ret[dic["ID"]] = {
                "title": dic["title"],
                "from_prog": FROM_PROG[dic["from_prog"]],
                "anal_type": ANAL_TYPE[dic["anal_type"]],
                "process_type": PROCESS_TYPE.get(dic.get("process_type")),
                "integer_format": dic.get("integer_format"),
                "value": dic["value"],
                "notes": dic["notes"],
            }
            not_parsed = set(("attach_id", "study_id", "location_id")) & set(dic.keys())
            if not_parsed:
                logging.warning(f"data named {not_parsed} not implemented for now!")
        return ret


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)
